from .html import HtmlWriter
from .markdown import MarkdownWriter


def writer(output_format=None):
    if output_format is None:
        output_format = "html"
    
    return _writers[output_format]()


def formats():
    return _writers.keys()


_writers = {
    "html": HtmlWriter,
    "markdown": MarkdownWriter,
}



from parsimonious.grammar import Grammar

from .. import html_paths


def read_html_path(string):
    path_node = _grammar.parse(string)
    return read_html_path_node(path_node)


def read_html_path_node(path_node):
    if path_node.children:
        return _read_html_path_elements_node(path_node.children[0])
    else:
        return html_paths.empty


def _read_html_path_elements_node(path_node):
    elements = [
        _read_element_node(child)
        for child in _repeated_children_with_separator(path_node, has_whitespace=True)
    ]
    return html_paths.path(elements)


def _read_element_node(node):
    tag_names = _read_tag_names_node(node.children[0])
    class_names = _read_class_names_node(node.children[1])
    fresh = _read_fresh_node(node.children[2])
    return html_paths.element(tag_names, class_names=class_names, fresh=fresh)


def _read_tag_names_node(node):
    return [
        child.text
        for child in _repeated_children_with_separator(node, has_whitespace=False)
    ]


def _read_class_names_node(class_names_node):
    return [
        _read_class_name_node(node)
        for node in class_names_node.children
    ]


def _read_class_name_node(node):
    return node.children[1].text


def _read_fresh_node(node):
    return len(node.children) > 0


def _repeated_children_with_separator(node, has_whitespace):
    yield node.children[0]
    
    if has_whitespace:
        sequence_node_index = 3
    else:
        sequence_node_index = 1
    
    sequence_node = node.children[1]
    for child in sequence_node.children:
        yield child.children[sequence_node_index]


grammar_text = r"""

html_path = html_path_elements?

html_path_elements = element (whitespace* ">" whitespace* element)*

element = tag_names class_name* fresh?

tag_names = identifier ("|" identifier)*

class_name = "." identifier

fresh = ":fresh"

identifier = ~"[A-Z0-9]+"i

whitespace = ~"\s"*

"""

_grammar = Grammar(grammar_text)




from .. import documents


def read_document_xml_element(element,
        body_reader,
        notes=None):
    
    if notes is None:
        notes = []
    
    body_element = element.find_child("w:body")
    return body_reader.read_all(body_element.children) \
        .map(lambda children: documents.document(children, documents.notes(notes)))



from xml.etree import ElementTree

from ..zips import open_zip, update_zip


_style_map_path = "mammoth/style-map"
_style_map_absolute_path = "/" + _style_map_path
_relationships_path = "word/_rels/document.xml.rels"
_content_types_path = "[Content_Types].xml"


def write_style_map(fileobj, style_map):
    with open_zip(fileobj, "r") as zip_file:
        relationships_xml = _generate_relationships_xml(zip_file.read_str(_relationships_path))
        content_types_xml = _generate_content_types_xml(zip_file.read_str(_content_types_path))
    
    update_zip(fileobj, {
        _style_map_path: style_map.encode("utf8"),
        _relationships_path: relationships_xml,
        _content_types_path: content_types_xml,
    })

def _generate_relationships_xml(relationships_xml):
    schema = "http://schemas.zwobble.org/mammoth/style-map";
    relationships_uri = "http://schemas.openxmlformats.org/package/2006/relationships"
    relationship_element_name = "{" + relationships_uri + "}Relationship"
    
    relationships = ElementTree.fromstring(relationships_xml)
    _add_or_update_element(relationships, relationship_element_name, "Id", {
        "Id": "rMammothStyleMap",
        "Type": schema,
        "Target": _style_map_absolute_path,
    })

    return ElementTree.tostring(relationships, "UTF-8")


def _generate_content_types_xml(content_types_xml):
    content_types_uri = "http://schemas.openxmlformats.org/package/2006/content-types"
    override_name = "{" + content_types_uri + "}Override"
    
    types = ElementTree.fromstring(content_types_xml)
    _add_or_update_element(types, override_name, "PartName", {
        "PartName": _style_map_absolute_path,
        "ContentType": "text/prs.mammoth.style-map",
    })
    
    return ElementTree.tostring(types, "UTF-8")


def _add_or_update_element(parent, name, identifying_attribute, attributes):
    existing_child = _find_child(parent, name, identifying_attribute, attributes)
    if existing_child is None:
        ElementTree.SubElement(parent, name, attributes)
    else:
        existing_child.attrib = attributes
    

def _find_child(parent, name, identifying_attribute, attributes):
    for element in parent.getiterator():
        if element.tag == name and element.get(identifying_attribute) == attributes.get(identifying_attribute):
            return element


def read_style_map(fileobj):
    with open_zip(fileobj, "r") as zip_file:
        if zip_file.exists(_style_map_path):
            return zip_file.read_str(_style_map_path)





from . import docx, conversion, options, images, underline
from .raw_text import extract_raw_text_from_element
from .docx.style_map import write_style_map, read_style_map

__all__ = ["convert_to_html", "extract_raw_text", "images", "underline"]


def convert_to_html(*args, **kwargs):
    return convert(*args, output_format="html", **kwargs)


def convert_to_markdown(*args, **kwargs):
    return convert(*args, output_format="markdown", **kwargs)


def convert(fileobj, transform_document=None, id_prefix=None, **kwargs):
    if transform_document is None:
        transform_document = lambda x: x
    if kwargs.get("style_map") is None:
        kwargs["style_map"] = read_style_map(fileobj)
    return options.read_options(kwargs).bind(lambda convert_options:
        docx.read(fileobj).map(transform_document).bind(lambda document:
            conversion.convert_document_element_to_html(
                document,
                id_prefix=id_prefix,
                **convert_options
            )
        )
    )
    

def extract_raw_text(fileobj):
    return docx.read(fileobj).map(extract_raw_text_from_element)


def embed_style_map(fileobj, style_map):
    write_style_map(fileobj, style_map)

def read_embedded_style_map(fileobj):
    return read_style_map(fileobj)



def read_content_types_xml_element(element):
    extension_defaults = dict(map(
        _read_default,
        element.find_children("content-types:Default")
    ))
    overrides = dict(map(
        _read_override,
        element.find_children("content-types:Override")
    ))
    return _ContentTypes(extension_defaults, overrides)


def _read_default(element):
    extension = element.attributes["Extension"]
    content_type = element.attributes["ContentType"]
    return extension, content_type


def _read_override(element):
    part_name = element.attributes["PartName"]
    content_type = element.attributes["ContentType"]
    return part_name.lstrip("/"), content_type


class _ContentTypes(object):
    _image_content_types = {
        "png": "png",
        "gif": "gif",
        "jpeg": "jpeg",
        "jpg": "jpeg",
        "tif": "tiff",
        "tiff": "tiff",
        "bmp": "bmp",
    }
    
    def __init__(self, extension_defaults, overrides):
        self._extension_defaults = extension_defaults
        self._overrides = overrides
    
    def find_content_type(self, path):
        if path in self._overrides:
            return self._overrides[path]

        extension = _get_extension(path)
        default_type = self._extension_defaults.get(extension)
        if default_type is not None:
            return default_type

        image_type = self._image_content_types.get(extension.lower())
        if image_type is not None:
            return "image/" + image_type
        
        return None


def _get_extension(path):
    return path.rpartition(".")[2]



# coding=utf-8

from __future__ import unicode_literals

import base64
import random

from . import documents, results, html_paths, images, writers
from .html_generation import HtmlGenerator, satisfy_html_path, append_html_path
from .docx.files import InvalidFileReferenceError


def convert_document_element_to_html(element,
        style_map=None,
        convert_image=None,
        convert_underline=None,
        id_prefix=None,
        output_format=None,
        ignore_empty_paragraphs=True):
            
    if style_map is None:
        style_map = []
    
    if id_prefix is None:
        id_prefix = str(random.randint(0, 1000000000000000))
    
    if convert_image is None:
        convert_image = images.inline(_generate_image_attributes)
    
    def create_writer():
        return writers.writer(output_format)
    
    html_generator = HtmlGenerator(create_writer)
    messages = []
    converter = _DocumentConverter(
        messages=messages,
        style_map=style_map,
        convert_image=convert_image,
        convert_underline=convert_underline,
        id_prefix=id_prefix,
        ignore_empty_paragraphs=ignore_empty_paragraphs,
        html_generator=html_generator,
        note_references=[])
    converter.visit(element)
    html_generator.end_all()
    return results.Result(html_generator.as_string(), messages)
    
    
def _generate_image_attributes(image):
    with image.open() as image_bytes:
        encoded_src = base64.b64encode(image_bytes.read()).decode("ascii")
    
    return {
        "src": "data:{0};base64,{1}".format(image.content_type, encoded_src)
    }


class _DocumentConverter(documents.ElementVisitor):
    def __init__(self, messages, style_map, convert_image, convert_underline, id_prefix, ignore_empty_paragraphs, html_generator, note_references):
        self._messages = messages
        self._style_map = style_map
        self._id_prefix = id_prefix
        self._ignore_empty_paragraphs = ignore_empty_paragraphs
        self._note_references = note_references
        self._convert_underline = convert_underline or self._default_convert_underline
        self._convert_image = convert_image
        self._html_generator = html_generator
    
    def _with_html_generator(self, html_generator):
        return _DocumentConverter(
            messages=self._messages,
            style_map=self._style_map,
            convert_image=self._convert_image,
            convert_underline=self._convert_underline,
            id_prefix=self._id_prefix,
            ignore_empty_paragraphs=self._ignore_empty_paragraphs,
            html_generator=html_generator,
            note_references=self._note_references
        )

    def visit_image(self, image):
        try:
            self._convert_image(image, self._html_generator)
        except InvalidFileReferenceError as error:
            self._messages.append(results.warning(str(error)))

    def visit_document(self, document):
        self._visit_all(document.children)
        self._html_generator.end_all()
        self._html_generator.start("ol")
        notes = [
            document.notes.resolve(reference)
            for reference in self._note_references
        ]
        self._visit_all(notes)
        self._html_generator.end()


    def visit_paragraph(self, paragraph):
        html_path = self._find_html_path_for_paragraph(paragraph)
        satisfy_html_path(self._html_generator, html_path)
        if not self._ignore_empty_paragraphs:
            self._html_generator.write_all()
        self._visit_all(paragraph.children)


    def visit_run(self, run):
        run_generator = self._html_generator.child()
        html_path = self._find_html_path_for_run(run)
        if html_path:
            satisfy_html_path(run_generator, html_path)
        if run.is_bold:
            run_generator.start("strong")
        if run.is_italic:
            run_generator.start("em")
        if run.vertical_alignment == documents.VerticalAlignment.superscript:
            run_generator.start("sup")
        if run.vertical_alignment == documents.VerticalAlignment.subscript:
            run_generator.start("sub")
        if run.is_underline:
            self._convert_underline(run_generator)
        if run.is_strikethrough:
            self._convert_strikethrough(run_generator)
        self._with_html_generator(run_generator)._visit_all(run.children)
        run_generator.end_all()
        self._html_generator.append(run_generator)
    
    
    def _default_convert_underline(self, run_generator):
        style = self._find_style(None, "underline")
        if style is not None:
            append_html_path(run_generator, style.html_path)
    
    
    def _convert_strikethrough(self, run_generator):
        style = self._find_style(None, "strikethrough")
        if style is None:
            run_generator.start("s")
        else:
            append_html_path(run_generator, style.html_path)
        
    

    def visit_text(self, text):
        self._html_generator.text(text.value)
    
    
    def visit_hyperlink(self, hyperlink):
        if hyperlink.anchor is None:
            href = hyperlink.href
        else:
            href = "#{0}".format(self._html_id(hyperlink.anchor))
        self._html_generator.start("a", {"href": href})
        self._visit_all(hyperlink.children)
        self._html_generator.end()
    
    
    def visit_bookmark(self, bookmark):
        self._html_generator.start("a", {"id": self._html_id(bookmark.name)}, always_write=True)
        self._html_generator.end()
    
    
    def visit_tab(self, tab):
        self._html_generator.text("\t")
    
    
    def visit_table(self, table):
        self._html_generator.end_all()
        self._html_generator.start("table")
        self._visit_all(table.children)
        self._html_generator.end()
    
    
    def visit_table_row(self, table_row):
        self._html_generator.start("tr")
        self._visit_all(table_row.children)
        self._html_generator.end()
    
    
    def visit_table_cell(self, table_cell):
        self._html_generator.start("td", always_write=True)
        for child in table_cell.children:
            child_generator = self._html_generator.child()
            self._with_html_generator(child_generator).visit(child)
            child_generator.end_all()
            self._html_generator.append(child_generator)
            
        self._html_generator.end()
    
    
    def visit_line_break(self, line_break):
        self._html_generator.self_closing("br")
    
    def visit_note_reference(self, note_reference):
        self._html_generator.start("sup")
        self._html_generator.start("a", {
            "href": "#" + self._note_html_id(note_reference),
            "id": self._note_ref_html_id(note_reference),
        })
        self._note_references.append(note_reference);
        note_number = len(self._note_references)
        self._html_generator.text("[{0}]".format(note_number))
        self._html_generator.end()
        self._html_generator.end()
    
    def visit_note(self, note):
        self._html_generator.start("li", {"id": self._note_html_id(note)})
        note_generator = self._html_generator.child()
        self._with_html_generator(note_generator)._visit_all(note.body)
        note_generator.text(" ")
        note_generator.start("a", {"href": "#" + self._note_ref_html_id(note)})
        note_generator.text(_up_arrow)
        note_generator.end_all()
        self._html_generator.append(note_generator)
        self._html_generator.end()


    def _visit_all(self, elements):
        for element in elements:
            self.visit(element)


    def _find_html_path_for_paragraph(self, paragraph):
        default = html_paths.path([html_paths.element("p", fresh=True)])
        return self._find_html_path(paragraph, "paragraph", default)
    
    def _find_html_path_for_run(self, run):
        return self._find_html_path(run, "run", default=None)
        
    
    def _find_html_path(self, element, element_type, default):
        style = self._find_style(element, element_type)
        if style is not None:
            return style.html_path
        
        if element.style_id is not None:
            self._messages.append(results.warning(
                "Unrecognised {0} style: {1} (Style ID: {2})".format(
                    element_type, element.style_name, element.style_id)
            ))
        
        return default
    
    def _find_style(self, element, element_type):
        for style in self._style_map:
            document_matcher = style.document_matcher
            if _document_matcher_matches(document_matcher, element, element_type):
                return style

    def _note_html_id(self, note):
        return self._html_id("{0}-{1}".format(note.note_type, note.note_id))
        
    def _note_ref_html_id(self, note):
        return self._html_id("{0}-ref-{1}".format(note.note_type, note.note_id))
    
    def _html_id(self, suffix):
        return "{0}-{1}".format(self._id_prefix, suffix)
        

def _document_matcher_matches(matcher, element, element_type):
    if matcher.element_type in ["underline", "strikethrough"]:
        return matcher.element_type == element_type
    else:
        return (
            matcher.element_type == element_type and (
                matcher.style_id is None or
                matcher.style_id == element.style_id
            ) and (
                matcher.style_name is None or
                element.style_name is not None and (matcher.style_name.upper() == element.style_name.upper())
            ) and (
                element_type != "paragraph" or
                matcher.numbering is None or
                matcher.numbering == element.numbering
            )
        )

_up_arrow = "↑"



import os
import contextlib
try:
    from urllib2 import urlopen
except ImportError:
    from urllib.request import urlopen
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


class Files(object):
    def __init__(self, base):
        self._base = base
    
    def open(self, uri):
        try:
            if _is_absolute(uri):
                return contextlib.closing(urlopen(uri))
            elif self._base is not None:
                return open(os.path.join(self._base, uri), "rb")
            else:
                raise InvalidFileReferenceError("could not find external image '{0}', fileobj has no name".format(uri))
        except IOError as error:
            message = "could not open external image: '{0}' (document directory: '{1}')\n{2}".format(
                uri, self._base, str(error))
            raise InvalidFileReferenceError(message)


def _is_absolute(url):
    return urlparse(url).scheme != ""


class InvalidFileReferenceError(ValueError):
    pass



from . import style_reader, lists, results


def read_options(options):
    custom_style_map_result = _read_style_map(options.get("style_map") or "")
    include_default_style_map = options.get("include_default_style_map", True)
    options["ignore_empty_paragraphs"] = options.get("ignore_empty_paragraphs", True)
    
    options["style_map"] = custom_style_map_result.value + \
        (_default_style_map if include_default_style_map else [])
    return custom_style_map_result.map(lambda _: options)


def _read_style_map(style_text):
    lines = filter(None, map(_get_line, style_text.split("\n")))
    return results.combine(lists.map(style_reader.read_style, lines)) \
        .map(lambda style_mappings: lists.filter(None, style_mappings))
    

def _get_line(line):
    line = line.strip()
    if line.startswith("#"):
        return None
    else:
        return line


_default_style_map_result = _read_style_map("""
p[style-name='Normal'] => p:fresh

p.Heading1 => h1:fresh
p.Heading2 => h2:fresh
p.Heading3 => h3:fresh
p.Heading4 => h4:fresh
p[style-name='Heading 1'] => h1:fresh
p[style-name='Heading 2'] => h2:fresh
p[style-name='Heading 3'] => h3:fresh
p[style-name='Heading 4'] => h4:fresh
p[style-name='heading 1'] => h1:fresh
p[style-name='heading 2'] => h2:fresh
p[style-name='heading 3'] => h3:fresh
p[style-name='heading 4'] => h4:fresh
p[style-name='heading 4'] => h4:fresh

r[style-name='Strong'] => strong

p[style-name='footnote text'] => p
r[style-name='footnote reference'] =>
p[style-name='endnote text'] => p
r[style-name='endnote reference'] =>
        
# LibreOffice
p[style-name='Footnote'] => p
r[style-name='Footnote anchor'] =>
p[style-name='Endnote'] => p
r[style-name='Endnote anchor'] =>

p:unordered-list(1) => ul > li:fresh
p:unordered-list(2) => ul|ol > li > ul > li:fresh
p:unordered-list(3) => ul|ol > li > ul|ol > li > ul > li:fresh
p:unordered-list(4) => ul|ol > li > ul|ol > li > ul|ol > li > ul > li:fresh
p:unordered-list(5) => ul|ol > li > ul|ol > li > ul|ol > li > ul|ol > li > ul > li:fresh
p:ordered-list(1) => ol > li:fresh
p:ordered-list(2) => ul|ol > li > ol > li:fresh
p:ordered-list(3) => ul|ol > li > ul|ol > li > ol > li:fresh
p:ordered-list(4) => ul|ol > li > ul|ol > li > ul|ol > li > ol > li:fresh
p:ordered-list(5) => ul|ol > li > ul|ol > li > ul|ol > li > ul|ol > li > ol > li:fresh

r[style-name='Hyperlink'] =>
""")


assert not _default_style_map_result.messages
_default_style_map = _default_style_map_result.value



import functools

from .. import lists
from .. import documents
from .. import results


def create_reader(note_type, body_reader):
    def read_notes_xml_element(element):
        note_elements = lists.filter(
            _is_note_element,
            element.find_children("w:" + note_type),
        )
        return results.combine(lists.map(_read_note_element, note_elements))


    def _is_note_element(element):
        return element.attributes.get("w:type") not in ["continuationSeparator", "separator"]


    def _read_note_element(element):
        return body_reader.read_all(element.children).map(lambda body: 
            documents.note(
                note_type=note_type,
                note_id=element.attributes["w:id"],
                body=body
            ))
    
    return read_notes_xml_element

create_footnotes_reader = functools.partial(create_reader, "footnote")
create_endnotes_reader = functools.partial(create_reader, "endnote")



import sys


def collect(values):
    result = []
    _collect(values, result)
    return result


def _collect(values, result):
    for value in values:
        if isinstance(value, list):
            _collect(value, result)
        elif value is not None:
            result.append(value)

def unique(values):
    output = []
    seen = set()
    for value in values:
        if value not in seen:
            seen.add(value)
            output.append(value)
    return output


def flat_map(func, values):
    return [
        element
        for value in values
        for element in func(value)
    ]


if sys.version_info[0] == 2:
    map = map
    filter = filter
else:
    import builtins
    def map(*args, **kwargs):
        return list(builtins.map(*args, **kwargs))
    def filter(*args, **kwargs):
        return list(builtins.filter(*args, **kwargs))



def inline(func):
    def convert_image(image, html_generator):
        attributes = func(image).copy()
        if image.alt_text:
            attributes["alt"] = image.alt_text
            
        html_generator.self_closing("img", attributes)
    
    return convert_image



from ..documents import numbering_level


def read_numbering_xml_element(element):
    abstract_nums = _read_abstract_nums(element)
    nums = _read_nums(element, abstract_nums)
    return Numbering(nums)


def _read_abstract_nums(element):
    abstract_num_elements = element.find_children("w:abstractNum")
    return dict(map(_read_abstract_num, abstract_num_elements))


def _read_abstract_num(element):
    abstract_num_id = element.attributes.get("w:abstractNumId")
    levels = _read_abstract_num_levels(element)
    return abstract_num_id, levels


def _read_abstract_num_levels(element):
    levels = map(_read_abstract_num_level, element.find_children("w:lvl"))
    return dict(
        (level.level_index, level)
        for level in levels
    )


def _read_abstract_num_level(element):
    level_index = element.attributes["w:ilvl"]
    num_fmt = element.find_child_or_null("w:numFmt").attributes.get("w:val")
    is_ordered = num_fmt != "bullet"
    return numbering_level(level_index, is_ordered)


def _read_nums(element, abstract_nums):
    num_elements = element.find_children("w:num")
    return dict(
        _read_num(num_element, abstract_nums)
        for num_element in num_elements
    )


def _read_num(element, abstract_nums):
    num_id = element.attributes.get("w:numId")
    abstract_num_id = element.find_child_or_null("w:abstractNumId").attributes["w:val"]
    return num_id, abstract_nums[abstract_num_id]


class Numbering(object):
    def __init__(self, nums):
        self._nums = nums
    
    def find_level(self, num_id, level):
        num = self._nums.get(num_id)
        if num is None:
            return None
        else:
            return num[level]



import contextlib
import io
import shutil

from zipfile import ZipFile


def open_zip(fileobj, mode):
    return _Zip(ZipFile(fileobj, mode))


class _Zip(object):
    def __init__(self, zip_file):
        self._zip_file = zip_file
    
    def __enter__(self):
        return self
    
    def __exit__(self, *args):
        self._zip_file.close()

    def open(self, name):
        return contextlib.closing(self._zip_file.open(name))

    def exists(self, name):
        try:
            self._zip_file.getinfo(name)
            return True
        except KeyError:
            return False

    def read_str(self, name):
        return self._zip_file.read(name).decode("utf8")


def update_zip(fileobj, files):
    source = ZipFile(fileobj, "r")
    try:
        destination_fileobj = io.BytesIO()
        destination = ZipFile(destination_fileobj, "w")
        try:
            names = set(source.namelist()) | set(files.keys())
            for name in names:
                if name in files:
                    contents = files[name]
                else:
                    contents = source.read(name)
                destination.writestr(name, contents)
        finally:
            destination.close()
    finally:
        source.close()
    
    fileobj.seek(0)
    destination_fileobj.seek(0)
    shutil.copyfileobj(destination_fileobj, fileobj)



from __future__ import unicode_literals


class HtmlGenerator(object):
    def __init__(self, create_writer):
        self._stack = []
        self._create_writer = create_writer
        self._writer = create_writer()
    
    def text(self, text):
        if text:
            self.write_all()
            self._writer.text(text)
    
    def start(self, name, attributes=None, always_write=False):
        self._stack.append(_Element(name, attributes))
        
        if always_write:
            self.write_all()

    def end(self):
        element = self._stack.pop()
        if element.written:
            self._writer.end(element.name)
    
    def end_all(self):
        while self._stack:
            self.end()
    
    def self_closing(self, name, attributes=None):
        self._writer.self_closing(name, attributes)
    
    def write_all(self):
        for element in self._stack:
            if not element.written:
                element.written = True
                self._write_element(element)
    
    def _write_element(self, element):
        self._writer.start(element.name, element.attributes)
    
    def append(self, other):
        other_string = other.as_string()
        if other_string:
            self.write_all()
            self._writer.append(other_string)
    
    def as_string(self):
        return self._writer.as_string()
    
    def child(self):
        return HtmlGenerator(self._create_writer)


class _Element(object):
    def __init__(self, name, attributes):
        if attributes is None:
            attributes = {}
        
        self.name = name
        self.attributes = attributes
        self.written = False


def satisfy_html_path(generator, path):
    first_unsatisfied_index = _find_first_unsatisfied_index(generator, path)
    while len(generator._stack) > first_unsatisfied_index:
        generator.end()
    
    for element in path.elements[first_unsatisfied_index:]:
        _append_element(generator, element)


def append_html_path(generator, path):
    for element in path.elements:
        _append_element(generator, element)


def _append_element(generator, element):
    attributes = {}
    if element.class_names:
        attributes["class"] = _generate_class_attribute(element)
    generator.start(element.names[0], attributes=attributes)
    


def _find_first_unsatisfied_index(generator, path):
    for index, (generated_element, path_element) in enumerate(zip(generator._stack, path.elements)):
        if not _is_element_match(generated_element, path_element):
            return index
    
    return len(generator._stack)


def _is_element_match(generated_element, path_element):
    return (
        not path_element.fresh and
        generated_element.name in path_element.names and
        generated_element.attributes.get("class", "") == _generate_class_attribute(path_element)
    )


def _generate_class_attribute(path_element):
    return " ".join(path_element.class_names)



import os

from .. import results, lists
from .document_xml import read_document_xml_element
from .content_types_xml import read_content_types_xml_element
from .relationships_xml import read_relationships_xml_element, Relationships
from .numbering_xml import read_numbering_xml_element, Numbering
from .styles_xml import read_styles_xml_element
from .notes_xml import create_footnotes_reader, create_endnotes_reader
from .files import Files
from . import body_xml, office_xml
from ..zips import open_zip


def read(fileobj):
    zip_file = open_zip(fileobj, "r")
    body_readers = _body_readers(getattr(fileobj, "name", None), zip_file)
    
    return _read_notes(zip_file, body_readers).bind(lambda notes:
        _read_document(zip_file, body_readers, notes))


def _read_notes(zip_file, body_readers):
    empty_result = results.success([])
    
    read_footnotes_xml = create_footnotes_reader(body_readers("footnotes"))
    footnotes = _try_read_entry_or_default(
        zip_file, "word/footnotes.xml", read_footnotes_xml, default=empty_result)
    
    read_endnotes_xml = create_endnotes_reader(body_readers("endnotes"))
    endnotes = _try_read_entry_or_default(
        zip_file, "word/endnotes.xml", read_endnotes_xml, default=empty_result)
    
    return results.combine([footnotes, endnotes]).map(lists.collect)
    
def _read_document(zip_file, body_readers, notes):
    with zip_file.open("word/document.xml") as document_fileobj:
        document_xml = office_xml.read(document_fileobj)
        return read_document_xml_element(
            document_xml,
            body_reader=body_readers("document"),
            notes=notes,
        )


def _body_readers(document_path, zip_file):
    with zip_file.open("[Content_Types].xml") as content_types_fileobj:
        content_types = read_content_types_xml_element(office_xml.read(content_types_fileobj))

    numbering = _try_read_entry_or_default(
        zip_file, "word/numbering.xml", read_numbering_xml_element, default=Numbering({}))
    
    with zip_file.open("word/styles.xml") as styles_fileobj:
        styles = read_styles_xml_element(office_xml.read(styles_fileobj))
    
    def for_name(name):
        relationships_path = "word/_rels/{0}.xml.rels".format(name)
        relationships = _try_read_entry_or_default(
            zip_file, relationships_path, read_relationships_xml_element,
            default=Relationships({}))
            
        return body_xml.reader(
            numbering=numbering,
            content_types=content_types,
            relationships=relationships,
            styles=styles,
            docx_file=zip_file,
            files=Files(None if document_path is None else os.path.dirname(document_path)),
        )
    
    return for_name


def _try_read_entry_or_default(zip_file, name, reader, default):
    if zip_file.exists(name):
        with zip_file.open(name) as fileobj:
            return reader(office_xml.read(fileobj))
    else:
        return default



from __future__ import unicode_literals

from .abc import Writer

import cgi


class HtmlWriter(Writer):
    def __init__(self):
        self._fragments = []
    
    def text(self, text):
        self._fragments.append(_escape_html(text))
    
    def start(self, name, attributes=None):
        attribute_string = _generate_attribute_string(attributes)
        self._fragments.append("<{0}{1}>".format(name, attribute_string))

    def end(self, name):
        self._fragments.append("</{0}>".format(name))
    
    def self_closing(self, name, attributes=None):
        attribute_string = _generate_attribute_string(attributes)
        self._fragments.append("<{0}{1} />".format(name, attribute_string))
    
    def append(self, html):
        self._fragments.append(html)
    
    def as_string(self):
        return "".join(self._fragments)


def _escape_html(text):
    return cgi.escape(text, quote=True)


def _generate_attribute_string(attributes):
    if attributes is None:
        return ""
    else:
        return "".join(
            ' {0}="{1}"'.format(key, _escape_html(attributes[key]))
            for key in sorted(attributes)
        )



import contextlib

from .. import documents
from .. import results
from .. import lists
from .xmlparser import node_types


def reader(numbering=None,
        content_types=None,
        relationships=None,
        styles=None,
        docx_file=None,
        files=None):
    
    read, read_all = _create_reader(
        numbering=numbering,
        content_types=content_types,
        relationships=relationships,
        styles=styles,
        docx_file=docx_file,
        files=files,
    )
    return _BodyReader(read, read_all)



class _BodyReader(object):
    def __init__(self, read, read_all):
        self._read = read
        self._read_all = read_all
    
    def read(self, element):
        result = self._read(element)
        return results.Result(result.value, result.messages)
    
    def read_all(self, elements):
        result = self._read_all(elements)
        return results.Result(result.value, result.messages)


def _create_reader(numbering, content_types, relationships, styles, docx_file, files):
    _ignored_elements = set([
        "office-word:wrap",
        "v:shadow",
        "v:shapetype",
        "w:bookmarkStart",
        "w:bookmarkEnd",
        "w:sectPr",
        "w:proofErr",
        "w:lastRenderedPageBreak",
        "w:commentRangeStart",
        "w:commentRangeEnd",
        "w:commentReference",
        "w:del",
        "w:footnoteRef",
        "w:endnoteRef",
        "w:pPr",
        "w:rPr",
        "w:tblPr",
        "w:tblGrid",
        "w:tcPr",
    ])

    def text(element):
        return _success(documents.Text(_inner_text(element)))


    def run(element):
        properties = element.find_child_or_null("w:rPr")
        style_id = properties \
            .find_child_or_null("w:rStyle") \
            .attributes.get("w:val")
        
        if style_id is None:
            style_name = None
        else:
            style_name = styles.find_character_style_by_id(style_id).name
        
        vertical_alignment = properties \
            .find_child_or_null("w:vertAlign") \
            .attributes.get("w:val")
        
        is_bold = properties.find_child("w:b")
        is_italic = properties.find_child("w:i")
        is_underline = properties.find_child("w:u")
        is_strikethrough = properties.find_child("w:strike")
        
        return _read_xml_elements(element.children) \
            .map(lambda children: documents.run(
                children=children,
                style_id=style_id,
                style_name=style_name,
                is_bold=is_bold,
                is_italic=is_italic,
                is_underline=is_underline,
                is_strikethrough=is_strikethrough,
                vertical_alignment=vertical_alignment,
            ))


    def paragraph(element):
        properties = element.find_child_or_null("w:pPr")
        style_id = properties \
            .find_child_or_null("w:pStyle") \
            .attributes.get("w:val")
        
        if style_id is None:
            style_name = None
        else:
            style_name = styles.find_paragraph_style_by_id(style_id).name
        
        numbering = _read_numbering_properties(properties.find_child_or_null("w:numPr"))
        
        return _read_xml_elements(element.children) \
            .map(lambda children: documents.paragraph(
                children=children,
                style_id=style_id,
                style_name=style_name,
                numbering=numbering,
            )) \
            .append_extra()

    def _read_numbering_properties(element):
        num_id = element.find_child_or_null("w:numId").attributes.get("w:val")
        level_index = element.find_child_or_null("w:ilvl").attributes.get("w:val")
        if num_id is None or level_index is None:
            return None
        else:
            return numbering.find_level(num_id, level_index)


    def tab(element):
        return _success(documents.tab())
    
    
    def table(element):
        return _read_xml_elements(element.children) \
            .map(documents.table)
    
    
    def table_row(element):
        return _read_xml_elements(element.children) \
            .map(documents.table_row)
    
    
    def table_cell(element):
        return _read_xml_elements(element.children) \
            .map(documents.table_cell)
    
    
    def read_child_elements(element):
        return _read_xml_elements(element.children)
    
    
    def pict(element):
        return read_child_elements(element).to_extra()
    
    
    def hyperlink(element):
        relationship_id = element.attributes.get("r:id")
        anchor = element.attributes.get("w:anchor")
        children_result = _read_xml_elements(element.children)
        if relationship_id is not None:
            href = relationships[relationship_id].target
            return children_result.map(lambda children: documents.hyperlink(href=href, children=children))
        elif anchor is not None:
            return children_result.map(lambda children: documents.hyperlink(anchor=anchor, children=children))
        else:
            return children_result
    
    def bookmark_start(element):
        name = element.attributes.get("w:name")
        if name == "_GoBack":
            return _empty_result
        else:
            return _success(documents.bookmark(name))
    
    
    def br(element):
        break_type = element.attributes.get("w:type")
        if break_type:
            warning = results.warning("Unsupported break type: {0}".format(break_type))
            return _empty_result_with_messages([warning])
        else:
            return _success(documents.line_break())
    
    def inline(element):
        alt_text = element.find_child_or_null("wp:docPr").attributes.get("descr")
        blips = element.find_children("a:graphic") \
            .find_children("a:graphicData") \
            .find_children("pic:pic") \
            .find_children("pic:blipFill") \
            .find_children("a:blip")
        return _read_blips(blips, alt_text)
    
    def _read_blips(blips, alt_text):
        return _combine_results(map(lambda blip: _read_blip(blip, alt_text), blips))
    
    def _read_blip(element, alt_text):
        return _read_image(lambda: _find_blip_image(element), alt_text)
    
    def _read_image(find_image, alt_text):
        image_path, open_image = find_image()
        content_type = content_types.find_content_type(image_path)
        image = documents.image(alt_text=alt_text, content_type=content_type, open=open_image)
        
        if content_type in ["image/png", "image/gif", "image/jpeg", "image/svg+xml", "image/tiff"]:
            messages = []
        else:
            messages = [results.warning("Image of type {0} is unlikely to display in web browsers".format(content_type))]
            
        return _element_result_with_messages(image, messages)
    
    def _find_blip_image(element):
        embed_relationship_id = element.attributes.get("r:embed")
        link_relationship_id = element.attributes.get("r:link")
        if embed_relationship_id is not None:
            return _find_embedded_image(embed_relationship_id)
        elif link_relationship_id is not None:
            return _find_linked_image(link_relationship_id)
    
    def _find_embedded_image(relationship_id):
        image_path = "word/" + relationships[relationship_id].target.lstrip("/")
        
        def open_image():
            image_file = docx_file.open(image_path)
            if hasattr(image_file, "__exit__"):
                return image_file
            else:
                return contextlib.closing(image_file)
        
        return image_path, open_image
    
    
    def _find_linked_image(relationship_id):
        image_path = relationships[relationship_id].target
        
        def open_image():
            return files.open(image_path)
        
        return image_path, open_image
    
    def read_imagedata(element):
        title = element.attributes.get("o:title")
        return _read_image(lambda: _find_embedded_image(element.attributes["r:id"]), title)
    
    def note_reference_reader(note_type):
        def note_reference(element):
            return _success(documents.note_reference(note_type, element.attributes["w:id"]))
        
        return note_reference
    
    def alternate_content(element):
        return read_child_elements(element.find_child("mc:Fallback"))
    
    handlers = {
        "w:t": text,
        "w:r": run,
        "w:p": paragraph,
        "w:tab": tab,
        "w:tbl": table,
        "w:tr": table_row,
        "w:tc": table_cell,
        "w:ins": read_child_elements,
        "w:smartTag": read_child_elements,
        "w:drawing": read_child_elements,
        "v:roundrect": read_child_elements,
        "v:shape": read_child_elements,
        "v:textbox": read_child_elements,
        "w:txbxContent": read_child_elements,
        "w:pict": pict,
        "w:hyperlink": hyperlink,
        "w:bookmarkStart": bookmark_start,
        "w:br": br,
        "wp:inline": inline,
        "wp:anchor": inline,
        "v:imagedata": read_imagedata,
        "w:footnoteReference": note_reference_reader("footnote"),
        "w:endnoteReference": note_reference_reader("endnote"),
        "mc:AlternateContent": alternate_content,
    }
    
    def read(element):
        handler = handlers.get(element.name)
        if handler is None:
            if element.name not in _ignored_elements:
                warning = results.warning("An unrecognised element was ignored: {0}".format(element.name))
                return _empty_result_with_messages([warning])
            else:
                return _success(None)
        else:
            return handler(element)
        

    def _read_xml_elements(elements):
        return _combine_results(map(read, elements)) \
            .map(lambda values: lists.collect(values))
    
    return read, _read_xml_elements


def _find(predicate, iterable):
    for item in iterable:
        if predicate(item):
            return item


def _find_child(element, name):
    return _find(lambda child: child.name == name, element.children)


def _inner_text(node):
    if node.node_type == node_types.text:
        return node.value
    else:
        return "".join(_inner_text(child) for child in node.children)



class _ReadResult(object):
    def __init__(self, value, extra, messages):
        if extra is None:
            extra = []
        self._result = results.Result((value, extra), messages)
    
    @property
    def value(self):
        return self._result.value[0]
    
    @property
    def extra(self):
        return self._result.value[1]
    
    @property
    def messages(self):
        return self._result.messages
    
    def map(self, func):
        result = self._result.map(lambda value: func(value[0]))
        return _ReadResult(result.value, self.extra, result.messages)
    
    def to_extra(self):
        return _ReadResult(None, _concat(self.extra, self.value), self.messages)
    
    def append_extra(self):
        return _ReadResult(_concat(self.value, self.extra), None, self.messages)

def _success(element):
    return _ReadResult(element, None, [])

def _element_result_with_messages(element, messages):
    return _ReadResult(element, None, messages)

def _combine_results(read_results):
    combined = results.combine(result._result for result in read_results)
    if combined.value:
        value, extras = map(list, zip(*combined.value))
        extra = sum(extras, [])
    else:
        value, extra = [], None
    return _ReadResult(value, extra, combined.messages)

_empty_result = _success(None)

def _empty_result_with_messages(messages):
    return _ReadResult(None, None, messages)

def _concat(*values):
    valid_values = list(filter(None, values))
    if not valid_values:
        return None
    elif len(valid_values) == 1:
        return valid_values[0]
    else:
        return sum(list(map(_to_list, valid_values)), [])

def _to_list(value):
    if isinstance(value, list):
        return value
    else:
        return [value]



from . import documents


def extract_raw_text_from_element(element):
    if isinstance(element, documents.Text):
        return element.value
    else:
        text = "".join(map(extract_raw_text_from_element, getattr(element, "children", [])))
        if isinstance(element, documents.Paragraph):
            return text + "\n\n"
        else:
            return text



from ..lists import flat_map
from .xmlparser import parse_xml, XmlElement


_namespaces = [
    ("w", "http://schemas.openxmlformats.org/wordprocessingml/2006/main"),
    ("wp", "http://schemas.openxmlformats.org/drawingml/2006/wordprocessingDrawing"),
    ("a", "http://schemas.openxmlformats.org/drawingml/2006/main"),
    ("pic", "http://schemas.openxmlformats.org/drawingml/2006/picture"),
    ("content-types", "http://schemas.openxmlformats.org/package/2006/content-types"),
    ("r", "http://schemas.openxmlformats.org/officeDocument/2006/relationships"),
    ("relationships", "http://schemas.openxmlformats.org/package/2006/relationships"),
    ("v", "urn:schemas-microsoft-com:vml"),
    ("mc", "http://schemas.openxmlformats.org/markup-compatibility/2006"),
    ("office-word", "urn:schemas-microsoft-com:office:word"),
]


def read(fileobj):
    return _collapse_alternate_content(parse_xml(fileobj, _namespaces))[0]
    

def _collapse_alternate_content(node):
    if isinstance(node, XmlElement):
        if node.name == "mc:AlternateContent":
            return node.find_child("mc:Fallback").children
        else:
            node.children = flat_map(_collapse_alternate_content, node.children)
            return [node]
    else:
        return [node]



from __future__ import absolute_import

import abc


class Writer(object):
    __metaclass__ = abc.ABCMeta
    
    @abc.abstractmethod
    def text(self, text):
        pass
    
    @abc.abstractmethod
    def start(self, name, attributes=None):
        pass

    @abc.abstractmethod
    def end(self, name):
        pass
    
    @abc.abstractmethod
    def self_closing(self, name, attributes=None):
        pass
    
    @abc.abstractmethod
    def append(self, html):
        pass
    
    @abc.abstractmethod
    def as_string(self):
        pass



import cobble


class Element(object):
    pass


class HasChildren(Element):
    children = cobble.field()


@cobble.data
class Document(HasChildren):
    notes = cobble.field()

@cobble.data
class Paragraph(HasChildren):
    style_id = cobble.field()
    style_name = cobble.field()
    numbering = cobble.field()

@cobble.data
class Run(HasChildren):
    style_id = cobble.field()
    style_name = cobble.field()
    is_bold = cobble.field()
    is_italic = cobble.field()
    is_underline = cobble.field()
    is_strikethrough = cobble.field()
    vertical_alignment = cobble.field()

@cobble.data
class Text(Element):
    value = cobble.field()

@cobble.data
class Hyperlink(HasChildren):
    href = cobble.field()
    anchor = cobble.field()

@cobble.data
class Table(HasChildren):
    pass

@cobble.data
class TableRow(HasChildren):
    pass

@cobble.data
class TableCell(HasChildren):
    pass

@cobble.data
class LineBreak(Element):
    pass

@cobble.data
class Tab(Element):
    pass
    

@cobble.visitable
class Image(Element):
    def __init__(self, alt_text, content_type, open):
        self.alt_text = alt_text
        self.content_type = content_type
        self.open = open


def document(children, notes=None):
    if notes is None:
        notes = Notes({})
    return Document(children, notes)

def paragraph(children, style_id=None, style_name=None, numbering=None):
    return Paragraph(children, style_id, style_name, numbering)

def run(children, style_id=None, style_name=None, is_bold=None, is_italic=None, is_underline=None, is_strikethrough=None, vertical_alignment=None):
    if vertical_alignment is None:
        vertical_alignment = VerticalAlignment.baseline
    return Run(
        children, style_id, style_name, bool(is_bold), bool(is_italic),
        bool(is_underline), bool(is_strikethrough), vertical_alignment,
    )

class VerticalAlignment(object):
    baseline = "baseline"
    superscript = "superscript"
    subscript = "subscript"

text = Text

_tab = Tab()

def tab():
    return _tab


image = Image

def hyperlink(children, href=None, anchor=None):
    return Hyperlink(href=href, anchor=anchor, children=children)


@cobble.data
class Bookmark(Element):
    name = cobble.field()

bookmark = Bookmark
    

table = Table
table_row = TableRow
table_cell = TableCell
line_break = LineBreak

def numbering_level(level_index, is_ordered):
    return _NumberingLevel(str(level_index), bool(is_ordered))

@cobble.data
class _NumberingLevel(object):
    level_index = cobble.field()
    is_ordered = cobble.field()

@cobble.data
class Note(Element):
    note_type = cobble.field()
    note_id = cobble.field()
    body = cobble.field()


note = Note


class Notes(object):
    def __init__(self, notes):
        self._notes = notes
    
    def find_note(self, note_type, note_id):
        return self._notes[(note_type, note_id)]
    
    def resolve(self, reference):
        return self.find_note(reference.note_type, reference.note_id)
    
    def __eq__(self, other):
        return isinstance(other, Notes) and self._notes == other._notes

    def __ne__(self, other):
        return not (self == other)

def notes(notes_list):
    return Notes(dict(
        (_note_key(note), note)
        for note in notes_list
    ))
    
def _note_key(note):
    return (note.note_type, note.note_id)

@cobble.data
class NoteReference(Element):
    note_type = cobble.field()
    note_id = cobble.field()

note_reference = NoteReference


ElementVisitor = cobble.visitor(Element)



from __future__ import unicode_literals

from .abc import Writer

import re


class _WriterOutput(object):
    def __init__(self, start, end=None, generate_end=None, anchor_position=None):
        if generate_end is None:
            generate_end = _constant(end)
        
        self.start = start
        self.generate_end = generate_end
        self.anchor_position = anchor_position


def _constant(value):
    def get():
        return value
    
    return get


class _MarkdownState(object):
    def __init__(self):
        self._list_state_stack = []
        self.list_state = None
        self.list_item_has_closed = False
    
    def update_list_state(self, list_state):
        self._list_state_stack.append(self.list_state)
        self.list_state = list_state
    
    def pop_list_state(self):
        self.list_state = self._list_state_stack.pop()


class _MarkdownListState(object):
    def __init__(self, ordered, indentation):
        self.ordered = ordered
        self.count = 0
        self.indentation = indentation


def _symmetric_wrapped(end):
    return _Wrapped(end, end)


class _Wrapped(object):
    def __init__(self, start, end):
        self._start = start
        self._end = end
    
    def __call__(self, attributes, markdown_state):
        return _WriterOutput(self._start, self._end)


def _hyperlink(attributes, markdown_state):
    href = attributes.get("href", "")
    if href:
        return _WriterOutput(
            "[", "]({0})".format(href),
            anchor_position="before",
        )
    else:
        return _default_output


def _image(attributes, markdown_state):
    src = attributes.get("src", "")
    alt_text = attributes.get("alt", "")
    if src or alt_text:
        return _WriterOutput("![{0}]({1})".format(alt_text, src), "")
    else:
        return _default_output


def _list(ordered):
    def call(attributes, markdown_state):
        if markdown_state.list_state is None:
            start = ""
            end_text = "\n"
            indentation = 0
        else:
            start = "\n"
            end_text = ""
            indentation = markdown_state.list_state.indentation + 1
        
        def generate_end():
            markdown_state.pop_list_state()
            return end_text
        
        markdown_state.update_list_state(_MarkdownListState(
            ordered=ordered,
            indentation=indentation,
        ))
        
        return _WriterOutput(start, generate_end=generate_end)
    
    return call


def _list_item(attributes, markdown_state):
    markdown_state.list_item_has_closed = False
    
    list_state = markdown_state.list_state or _MarkdownListState(ordered=False, indentation=0)
    list_state.count += 1
    
    if list_state.ordered:
        bullet = "{0}.".format(list_state.count)
    else:
        bullet = "-"
    
    def generate_end():
        if markdown_state.list_item_has_closed:
            return ""
        else:
            markdown_state.list_item_has_closed = True
            return "\n"
    
    return _WriterOutput(
        start=("\t" * list_state.indentation) + bullet + " ",
        generate_end=generate_end
    )


def _init_writers():
    writers = {
        "p": _Wrapped("", "\n\n"),
        "br": _Wrapped("", "  \n"),
        "strong": _symmetric_wrapped("__"),
        "em": _symmetric_wrapped("*"),
        "a": _hyperlink,
        "img": _image,
        "ol": _list(ordered=True),
        "ul": _list(ordered=False),
        "li": _list_item,
    }
    
    for level in range(1, 7):
        writers["h{0}".format(level)] = _Wrapped("#" * level + " ", "\n\n")
    
    return writers


_writers = _init_writers()
_default_output = _WriterOutput("", "")

def _default_writer(attributes, markdown_state):
    return _default_output


class MarkdownWriter(Writer):
    def __init__(self):
        self._fragments = []
        self._element_stack = []
        self._markdown_state = _MarkdownState()
    
    def text(self, text):
        self._fragments.append(_escape_markdown(text))
    
    def start(self, name, attributes=None):
        if attributes is None:
            attributes = {}
        
        output = _writers.get(name, _default_writer)(attributes, self._markdown_state)
        self._element_stack.append(output.generate_end)
        
        anchor_before_start = output.anchor_position == "before"
        if anchor_before_start:
            self._write_anchor(attributes)
        
        self._fragments.append(output.start)
        
        if not anchor_before_start:
            self._write_anchor(attributes)
        
        

    def end(self, name):
        end = self._element_stack.pop()
        output = end()
        self._fragments.append(output)
    
    def self_closing(self, name, attributes=None):
        self.start(name, attributes)
        self.end(name)
    
    def append(self, other):
        self._fragments.append(other)
    
    def as_string(self):
        return "".join(self._fragments)
    
    def _write_anchor(self, attributes):
        html_id = attributes.get("id")
        if html_id:
            self._fragments.append('<a id="{0}"></a>'.format(html_id))


def _escape_markdown(value):
    return re.sub(r"([\`\*_\{\}\[\]\(\)\#\+\-\.\!])", r"\\\1", re.sub("\\\\", "\\\\\\\\", value))
